%******************************************************************************;
%* SAS macro Num2Char vs. 0.0.2 , 21 Oct 2013, by Jim Groeneveld, Netherlands *;
%******************************************************************************;
%* STILL UNDER CONSTRUCTION, NOT FOR VALIDATION, USE WITH CAUTION AT OWN RISK *;
%*                                                                            *;
   %MACRO   Num2Char   /* Numeric to character conversion, preserving attr.s  */
/* -------------------------------------------------------------------------- */
/* ---------- A l w a y s   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
(  Data    = _LAST_    /* Input dataset with variable(s) to convert         R */
,  Out     = _SAME_    /* Output dataset, _SAME_: (input) dataset           R */
,  OverWrit= No        /* Yes: overwrite existing &Out output dataset       R */
,  VarList =           /* Variable list of numerical values to convert      R */
/*                                                                            */
/* ---------- R a r e l y   s p e c i f i e d   a r g u m e n t s ----------- */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  ConvFmt = BEST32.   /* format to use for conversion into 32 chars. var.  R */
                       /*   e.g. BEST16. or any other user specified format   */
,  CharLen = 32 /*16*/ /* corresponding length of resulting char. variable  R */
,  Version = 0.0.2     /* forced version specification for version control  O */
,  CalledBy=           /* Macro name of macro or program for err0r report   O */
,  Presence=           /* if not empty value is returned checking existence O */
/*                                                                            */
/* ----------- N e v e r   s p e c i f i e d   a r g u m e n t s ------------ */
/* Argument Default       Description and remarks      Required/Optional: R/O */
/* -------- --------      --------------------------------------------------- */
,  Debug   = 0         /* Debug level; for testing, debugging purposes only O */
,  ForMyUse=           /* argument for the program's use only, don't use it O */
)/ DES     =              'Numeric to character variable conversion' /*Store*/ ;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* File       : Num2Char.SAS / Num2Char.002                                   *;
%*                                                                            *;
%* Purpose    : Numeric to character conversion, preserving var. attributes   *;
%*                                                                            *;
%* Input      : Input dataset with numeric variable(s) to convert             *;
%*                                                                            *;
%* Output     : Output dataset with character converted numeric variable(s)   *;
%*                                                                            *;
%* Programmer : Jim Groeneveld                                                *;
%* Date       : 21 Oct 2013                                                   *;
%* Version    : 0.0.2                                                         *;
%*                                                                            *;
%* SAS version: 9.3 under Windows 7                                           *;
%*                                                                            *;
%* Validation : -                                                             *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Download location                                                          *;
%* -----------------                                                          *;
%* http://jim.groeneveld.eu.tf/software/SASmacro/Num2Char.zip                 *;
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Example macro call                                                         *;
%* ------------------                                                         *;
%*                                                                            *;
%* Checking presence of macro in calling code:                                *;
%* - - - - - - - - - - - - - - - - - - - - - -                                *;
/* %LET Checking = %Num2Char(Presence=any text);                              */
/* %IF (&Checking NE any text) %THEN ......... (macro not present) .........; */
%* or:                                                                        *;
/* %LET Checking = %Num2Char(Presence=%NRSTR(&MacName._&MacVs);               */
/* %IF (&Checking NE Num2Char_0.0.2) %THEN ..... (MacroVs not present) .....; */
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* Used auxiliary macros:                                                     *;
%* ----------------------                                                     *;
/* = internal: %DefLabel                                                      */
/* = attached: -                                                              */
/* = external: %UstrMcmp, %ERR0R, %WARN1NG, %_Delete_                         */
%*                                                                            *;
%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -*;
%*                                                                            *;
%* History of Num2Char:                                                       *;
%* ====================                                                       *;
%* version  date         description of changes, additions, improvements      *;
%* -------  -----------  ---------------------------------------------------- *;
%* 0.0.0     8 Oct 2013  initial prototype version including support for SAS  *;
%*                         name literal variable names (with e.g. NLITERAL)   *;
%* 0.0.1    10 Oct 2013  added code to transfer original variable labels to   *;
%*                         recoded character variables (with macro DefLabel)  *;
%* 0.0.2    21 Oct 2013  fixed bug by DROPping original numeric variables     *;
%*                         only (instead of unselective wildcard spec.)       *;
%*                                                                            *;
%* Future of Num2Char:                                                        *;
%* ===================                                                        *;
%* - if VarList is empty or _ALL_ then convert _all_ numerical variables      *;
%* - optional transfer of (in)formats if corresponding character fmts exist   *;
%* - dataset options support (already experimentally programmed separately)   *;
%*                                                                            *;
%******************************************************************************;

  %LOCAL     /* Define internal macro variables explicitely local */
    MacName  /* This macros external name: MetaData */
    MacVs    /* Version number major.minor.patchSub of macro: 0.0.2 */
    MacDate  /* Date of this version of macro: 21 Oct 2012 */
    ErrCount /* number of erroneously specified arguments */
    ArgList  /* List of user specifiable macro arguments */
    Argument /* Single user specifiable macro argument */
    I        /* counter */
    ValidVar /* stored VALIDVARNAME option */
    NoMPrint /* stored MPRINT option */
    NoMErr0r /* stored MERR0R option */
    NoSErr0r /* stored SERR0R option */
    NoMLogic /* stored MLOGIC option */
    Symbolgn /* stored SYMBOLGEN option */
    MacroGen /* stored MACROGEN option */
    NoSpool  /* stored SPOOL option */
    DataLib  /* Library name of DataFile (for splitting Lib and Set) */
    DataSet  /* Dataset name of DataFile (for splitting Lib and Set) */
    VarNames /* macro variable list of variable names in dataset &Data */
    /*VarCount*/ /* number of variables in dataset &Data */
    oVarList /* original variable list in dataset in original order */
    Contents /* output dataset name of PROC CONTENTS */
    vVarList /* validated VarList with numerical variables only */
    VarListN /* number of variables specified in (validated) VarList */
  ;

  %LET MacName  = Num2Char;
  %LET MacVs    = 0.0.2 ;
  %LET MacDate  = 21 Oct 2012;
  %LET ValidVar = %SYSFUNC(GETOPTION(VALIDVARNAME));
  %LET NoMPrint = %SYSFUNC(GETOPTION(MPRINT));
  %LET NoMErr0r = %SYSFUNC(GETOPTION(MERROR));
  %LET NoSErr0r = %SYSFUNC(GETOPTION(SERROR));
  %LET NoMLogic = %SYSFUNC(GETOPTION(MLOGIC));
  %LET Symbolgn = %SYSFUNC(GETOPTION(SYMBOLGEN));
  %LET MacroGen = %SYSFUNC(GETOPTION(MACROGEN));
  %LET NoSpool  = %SYSFUNC(GETOPTION(SPOOL));
  %LET Contents = __Cont__; %* or uniquely generated;
  %LET ErrCount = 0; %* to start with;

  %IF (%LENGTH(%SUPERQ(Presence))) %THEN %DO;
    %UNQUOTE(&Presence) /* no semicolon, value returned */
    %GOTO Finish;
  %END;

  %IF ( %UPCASE(&ForMyUse) = COMPILE ) %THEN %GOTO Finish; %* Compilation;

  %IF (&Debug EQ 1 OR &Debug GT 11) %THEN %DO;
    OPTIONS MPRINT MERROR SERROR MLOGIC SYMBOLGEN MACROGEN SPOOL;
  %END;

%*~~~~~~~~~~~~~~~~~~~~~~~~~;
%* Macro header in LOG file;
%*_________________________;

  %PUT;
  %PUT |~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~|;
  %PUT |   Macro &MacName, vs. &MacVs, by Jim Groeneveld, &MacDate  |;
  %PUT |  Search specific values in many datasets in many libraries.  |;
  %PUT |______________________________________________________________|;
  %PUT;

%* Feedback of specified arguments, programmed as macro loop;
  %PUT &MacName: Review of argument specifications (incl. defaults);
  %LET ArgList = Data Out Overwrit VarList ConvFmt Version CalledBy Debug;
  %LET I = 1;
  %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %DO %WHILE (%LENGTH(&Argument));
    %PUT &MacName: &Argument = &&&Argument;
    %LET I = %EVAL ( &I + 1 );
    %LET Argument = %SCAN ( &ArgList, &I, %STR( ) );
  %END;

  %IF (&Version NE AND &Version NE &MacVs) %THEN
  %DO;
    %PUT *** &MacName *** %ERR0R-: Specified version &Version does not match &MacVs,;
    %PUT .                       macro &MacName will abort;
    %GOTO Exit;
  %END;

%* Process keyword _LAST_ with Data;
  %IF (%UPCASE(&Data) EQ _LAST_) %THEN %LET Data = &SYSLAST;
%* Process keyword _SAME_ with Out;
  %IF (%UPCASE(&Out) EQ _SAME_) %THEN %LET Out = &Data;

%* Argument checks;
%* Data;
  %IF (NOT %LENGTH(%SUPERQ(Data))) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Data specified empty but is required,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;
  %IF ( NOT %SYSFUNC(EXIST(&Data)) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Dataset Data=&Data does not exist,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;
%* (no check yet on the contents of Data (variables and so on));

%* Overwrit and Out;
  %IF (NOT %LENGTH(%SUPERQ(Out))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: Out specified empty but is required,;
    %PUT .                     Out set to default Num2Char (in WORK);
    %LET Out = Num2Char;
  %END;
  %IF ( NOT %UstrMcmp(&Overwrit,Yes) AND NOT %UstrMcmp(&Overwrit,No) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Overwrit argument not specified or illegal: &Overwrit,;
    %PUT .                      macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;
  %ELSE %IF ( %SYSFUNC(EXIST(&Out)) AND %UstrMcmp(&OverWrit,No) ) %THEN %DO;
    %PUT *** &MacName *** %ERR0R-: Dataset Out=&Out exists and OverWrit is No,;
    %PUT .                     macro &MacName will abort;
    %LET ErrCount = %EVAL ( &ErrCount + 1 );
  %END;

%* ConvFmt;
  %IF (NOT %LENGTH(%SUPERQ(ConvFmt))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: ConvFmt specified empty but is required,;
    %PUT .                     Out set to default BEST32.;
    %LET ConvFmt = BEST32.;
  %END;
  %ELSE %IF (NOT %SYSFUNC(INDEX(&ConvFmt,.))) %THEN %DO;
    %PUT *** &MacName *** %WARN1NG-: ConvFmt lacks required period,;
    %PUT .                     Period added to ConvFmt at end;
    %LET ConvFmt = &ConvFmt..;
  %END;

%* later: VarList = _ALL_ or empty: all numeric variables;
%* VarList;

  %IF ( &ErrCount GT 0 ) %THEN %GOTO Aborting;

%* - - - - - - - - - - - ;
%* Core code starts here ;
%* - - - - - - - - - - - ;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

%* Process multiple dataset specifications from keyword &Data ;
  %IF (%LENGTH ( %SCAN ( &Data, 2, . ))) %THEN %DO;
    %LET DataLib = %SCAN ( &Data, 1, . );
    %LET Dataset = %SCAN ( &Data, 2, . );
  %END;
  %ELSE %DO;
    %LET DataLib = WORK;
    %LET Dataset = %SCAN ( &Data, 1, . );
  %END;

* Save list of original variables in dataset in original order;
  PROC SQL NOPRINT;
    SELECT NLITERAL(NAME) INTO :VarNames SEPARATED BY ' '  /* variable name list */
      FROM SASHELP.VCOLUMN
      WHERE UPCASE(LIBNAME) EQ UPCASE("&DataLib")
        AND UPCASE(MemName) EQ UPCASE("&Dataset");
*** SELECT COUNT(*) INTO :VarCount               /* number of elements */
      FROM SASHELP.VCOLUMN
      WHERE UPCASE(LIBNAME) EQ UPCASE("&DataLib")
        AND UPCASE(MemName) EQ UPCASE("&Dataset");
  RUN; QUIT;

%* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

* Investigate contents of &Data;
  PROC CONTENTS DATA=&Data (KEEP=&VarList) OUT=&Contents NOPRINT; RUN;

* Check whether all specified variabes indeed are numeric;
  %LET vVarList = ; %* initialise empty validated varlist;
  DATA _NULL_;
    SET &Contents END=_LastRec;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* Start building macro DefLabel with variable label attributes       ;
* to call to define labels with recoded variables in next data step. ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
    IF (_N_ EQ 1) THEN CALL EXECUTE ('%NRSTR(%MACRO DefLabel;)');
    IF (Type EQ 1) THEN DO; * Type = 1, numeric;
      _Numeric + 1;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* Build validated numerical variable list without specified character variables. ; 
* Macro statement directly executed, not part of macro DefLabel (unless NRSTR).  ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
      CALL EXECUTE ('%LET vVarList = &vVarList ' || TRIM(NLITERAL(Name)) || ';');
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* This can be done using a SAS variable as well but it should have to be defined ;
* with a LENGTH of $32767 in order to store a large list of variable names.      ;
* At the end (_LastRec) the value could be SYMPUT to a macro variable.           ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* It can also be done using CALL SYMPUT only, extending a macro value likewise.  ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
* Store variable label attributes of original numerical variables in macro ;
* DefLabel. This is SAS code to be executed delayed as part of the macro.  ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -;
      CALL EXECUTE (' ' || TRIM(NLITERAL(NAME)) || '=' || QUOTE(TRIM(LABEL)));

* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* Other, more cumbersome, less preferred but rather obvious ways to do this are:;
* 1. Use a single macro variable value to store all variable label definitions  ;
*    in the form ['variable name'n='variable label']. A variable name may be 32 ;
*    characters long and a macro label 256. Hence the maximum langth of such a  ;
*    definition per variable (including a separation space) is 295 chaaracters. ;
*    As the maximum macro value length is 65543 characters the minimum number   ;
*    of variable label definitions is 222. This may be sufficient in most cases ;
*    but occasionally may appear insufficient. E.g. one macro variable used in: ;
*    PROC SQL; *                                                                ;
*      SELECT TRIM(NAME)||'='||QUOTE(TRIM(LABEL))                               ;
*        INTO :Labels SEPARATED BY ' '                                          ;
*          FROM DICTIONARY.COLUMNS ... etc.                                     ;
*    See [https://groups.google.com/forum/#!topic/comp.soft-sys.sas/jDKvxvHdvp4];
*    --> Drawback: limited, possibly too less macro value space.                ;
* 2. Use as many (numbered) storage macro variables as labels to define.        ;
*    This results in many macro variables that also have to be defined %LOCAL.  ;
*    --> Drawback: many macro variables and a macro loop to process them.       ;
* 3. Build a macro that is to be called from the location where the labels are  ;
*    to be interted by reading them from another dataset using macro functions. ;
*    Such a method is decribed in: "Dynamically Create Variable Labels from     ;
*    Data Set Values" [http://www.lexjansen.com/pharmasug/2008/cc/cc04.pdf] by  ;
*    Wei Xu. This way is copied by SAS: [http://support.sas.com/kb/26/139.html] ;
*    --> Drawback: good idea, but much more code than actually needed (here).   ;
* 4. Build code to define variable labels into a SAS or TXT file to %INClude.   ;
*    See [https://groups.google.com/forum/#!topic/comp.soft-sys.sas/jDKvxvHdvp4];
*    --> Drawback: temporary external file, location to choose and to delete.   ;
* 5. Store variable labels in hash table by variable, to be used just once each.;
*    --> Drawback: could need rather much memory for just one time use each.    ;
* 6. Read/loop through the Contents dataset performing several actions:         ;
*    IF (_N_ EQ 1) THEN CALL EXECUTE ('PROC DATASETS beginning statements...'); ;
*    CALL EXECUTE (' ' || TRIM(NAME) || '=' || QUOTE(TRIM(LABEL)));             ;
*    IF (last_record_processed) THEN CALL EXECUTE ('finish PROC DATASETS...');  ;
*    --> Drawback: good, but here more code than actually needed (here).        ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;
* So the solution chosen here, instead of storing all variable labels in one    ;
* macro variable value (ex. 1), is to store the variable labels in a _macro_,   ;
* defined via CALL EXECUTE and filled likewise (see code above and below).      ;
* There already is a data step reading the Contents dataset (defining the macro ;
* DefLabel) and another one with the recoded variables with the original names  ;
* where the definitions in the form of the macro call can be activated.         ;
* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - ;

    END;
    ELSE DO; * Type = 2, character;
      PUT "*** &MacName *** Note-: Variable " Name "already is character, ignored.";
    END;
    IF (_LastRec) THEN DO;
      CALL SYMPUT ('VarListN', PUT(_Numeric,BEST.));
* End building macro with variable label attributes to call with recoded variables;
      CALL EXECUTE ('%NRSTR(%MEND;)'); * end stored variable label definitions;
    END;
  RUN;

* Remove leading spaces from &VarListN;
  %LET VarListN = &VarListN;

  %if &debug %then %do;
    %put DEBUG: Definition of macro DefLabel = (variable label attributes);
    %put %DefLabel; * show stored variable label definitions;

    %put DEBUG: VarNames=&VarNames;
    %put DEBUG: VarListN=&VarListN;
    %put DEBUG: vVarList=&vVarList;
  %end;

  DATA &Out;
    RETAIN &VarNames; * original variable list in original order;
    SET &Data (RENAME=(
      %DO I = 1 %TO &VarListN;
        %SYSFUNC(SCANQ(&vVarList,&I,%STR( ))) = __&I._
      %END;
      ));
    ARRAY _____Num /*_NUMERIC_*/
      %DO I = 1 %TO &VarListN;
        __&I._
      %END;
      ;
    ARRAY ____Char $32 &vVarList;
    DO _I_ = 1 TO DIM(____Char); %* or &VarListN;
      ____Char[_I_] = PUT ( _____Num[_I_], &ConvFmt );
    END;
* Assign stored original numerical variable labels to recoded character variables;
    LABEL %DefLabel; * activate stored variable label definitions;
* Drop original numeric variables (only those from the varlist been converted);
    DROP
      %DO I = 1 %TO &VarListN;
        __&I._
      %END;
    ;
  RUN;

/*
data current.original; set &data; run; * DEBUG;
data current.num2char; set &out; run; * DEBUG;
*/

  %LET SYSLAST = &Out;

%Aborting:
%* - - - - - - ;
%* Cleaning up ;
%* - - - - - - ;

%* Restore possibly modified option settings to previous ones;
  %if (&Debug) %then %do;
    OPTIONS &NoMPrint &NoMErr0r &NoSErr0r &NoMLogic &Symbolgn &MacroGen &NoSpool;
  %end;
  OPTIONS VALIDVARNAME = &ValidVar;

%* Delete work datasets;
  %_Delete_ (&Contents);

%Exit:
%Finish:

%MEND Num2Char;